/* Copyright (c) 2009 The Regents of the Signet Group.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 *  Authors:
 *  Angelo P. Castellani <castellani@dei.unipd.it>
 *  Mattia Gheda <ghedamat@dei.unipd.it>
 */

/*
 * CoAP header
     0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Ver| T |  OC   |      Code     |          Message ID           |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

#ifdef i386
typedef struct  {
  nx_uint8_t oc:4;
  nx_uint8_t t:2;
  nx_uint8_t ver:2;
  nx_uint8_t code;
  nx_uint16_t mid;
} coap_header_t;
#else
typedef nx_struct  {
  nx_uint8_t ver:2;
  nx_uint8_t t:2;
  nx_uint8_t oc:4;
  nx_uint8_t code;
  nx_uint16_t mid;
} coap_header_t __attribute__ ((__packed__));
#endif

/*
 * Coap OPTION HEADER
     0   1   2   3   4   5   6   7
   +---+---+---+---+---+---+---+---+
   | Option Delta  |    Length     | for 0..14
   +---+---+---+---+---+---+---+---+
   |   Option Value ...
   +---+---+---+---+---+---+---+---+
                                               for 15..270:
   +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
   | Option Delta  | 1   1   1   1 |          Length - 15          |
   +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
   |   Option Value ...
   +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
*/

#ifdef i386
typedef struct {
  nx_uint8_t len:4;
  nx_uint8_t od:4; //option delta
} coap_option_header_t ;
#else
typedef nx_struct {
  nx_uint8_t od:4; //option delta
  nx_uint8_t len:4;
} coap_option_header_t __attribute__ ((__packed__));

#endif

/*
 * Coap OPTION TYPES
       +-----+-----+---------------+--------+---------+-------------+
      | No. | C/E | Name          | Format | Length  | Default     |
      +-----+-----+---------------+--------+---------+-------------+
      |   1 | C   | Content-Type  | uint   | 1-2 B   | 0           |
      |   2 | E   | Max-Age       | uint   | 0-4 B   | 60          |
      |   3 | C   | Proxy-Uri     | string | 1-270 B | (none)      |
      |   4 | E   | Etag          | opaque | 1-4 B   | (none)      |
      |   5 | C   | Uri-Host      | string | 1-270 B | (see below) |
      |   6 | E   | Location-Path | string | 1-270 B | (none)      |
      |   7 | C   | Uri-Port      | uint   | 0-2 B   | (see below) |
      |   9 | C   | Uri-Path      | string | 1-270 B | (none)      |
      |  11 | C   | Token         | opaque | 1-8 B   | (empty)     |
      |  15 | C   | Uri-Query     | string | 1-270 B | (none)      |
      +-----+-----+---------------+--------+---------+-------------+
*/
 
enum coap_option_types {
  COAP_OPTION_CONTENT_TYPE = 1,
  COAP_OPTION_MAX_AGE,
  COAP_OPTION_PROXY_URI,
  COAP_OPTION_ETAG,
  COAP_OPTION_URI_HOST,
  COAP_OPTION_LOCATION_PATH,
  COAP_OPTION_URI_PORT = 8,
  COAP_OPTION_URI_PATH = 9,
  COAP_OPTION_TOKEN = 11,
  COAP_OPTION_URI_QUERY = 15
};



enum coap_v_values {
  COAP_VERSION = 1
};

enum coap_t_values {
  COAP_T_CON = 0,
  COAP_T_NON,
  COAP_T_ACK,
  COAP_T_RST
};

/*
 * COAP METHOD VALUES
                     +------+--------+---------------+
                     | Code | Name   | Reference     |
                     +------+--------+---------------+
                     |    1 | GET    | Section 5.8.1 |
                     |    2 | POST   | Section 5.8.2 |
                     |    3 | PUT    | Section 5.8.3 |
                     |    4 | DELETE | Section 5.8.4 |
                     +------+--------+---------------+
*/
enum coap_method_values {
  COAP_METHOD_GET = 1,
  COAP_METHOD_POST,
  COAP_METHOD_PUT,
  COAP_METHOD_DELETE
  
};

enum coap_msgtype_values {
  COAP_MSGTYPE_REQUEST = 0,
  COAP_MSGTYPE_RESPONSE,
  COAP_MSGTYPE_EMPTY
};

// TODO: add other codes see Sec. 11.1 enum per gli status

/*
 * Coap STATUS VALUES
        +------+-------------------------------+-----------------+
        | Code | Description                   | Reference       |
        +------+-------------------------------+-----------------+
        |   64 | 2.00 OK                       | Section 5.9.1.1 |
        |   65 | 2.01 Created                  | Section 5.9.1.2 |
        |   66 | 2.02 Deleted                  | Section 5.9.1.3 |
        |   67 | 2.03 Valid                    | Section 5.9.1.4 |
        |   68 | 2.04 Changed                  | Section 5.9.1.5 |
        |  128 | 4.00 Bad Request              | Section 5.9.2.1 |
        |  129 | 4.01 Unauthorized             | Section 5.9.2.2 |
        |  130 | 4.02 Bad Option               | Section 5.9.2.3 |
        |  131 | 4.03 Forbidden                | Section 5.9.2.4 |
        |  132 | 4.04 Not Found                | Section 5.9.2.5 |
        |  133 | 4.05 Method Not Allowed       | Section 5.9.2.6 |
        |  141 | 4.13 Request Entity Too Large | Section 5.9.2.7 |
        |  143 | 4.15 Unsupported Media Type   | Section 5.9.2.8 |
        |  160 | 5.00 Internal Server Error    | Section 5.9.3.1 |
        |  161 | 5.01 Not Implemented          | Section 5.9.3.2 |
        |  162 | 5.02 Bad Gateway              | Section 5.9.3.3 |
        |  163 | 5.03 Service Unavailable      | Section 5.9.3.4 |
        |  164 | 5.04 Gateway Timeout          | Section 5.9.3.5 |
        |  165 | 5.05 Proxying Not Supported   | Section 5.9.3.6 |
        +------+-------------------------------+-----------------+
*/ 
enum coap_status_values {
  COAP_STATUS_OK = 64,
  COAP_STATUS_OK_CREATED,
  COAP_STATUS_OK_DELETED,
  COAP_STATUS_OK_VALID,
  COAP_STATUS_OK_CHANGED,
  COAP_STATUS_OK_CONTENT,
  // XXX to be removed
  COAP_STATUS_FAILURE = 128,

  COAP_STATUS_BAD_REQUEST = 128,
  COAP_STATUS_UNAUTHORIZED,
  COAP_STATUS_BAD_OPTION,
  COAP_STATUS_FORBIDDEN,
  COAP_STATUS_NOT_FOUND,
  COAP_STATUS_METHOD_NOT_ALLOWED,
  COAP_STATUS_REQUEST_ENTITY_TOO_LARGE =141,
  COAP_STATUS_UNSUPPORTED_MEDIA_TYPE =143,
  COAP_STATUS_INTERNAL_SERVER_ERROR =160,
  COAP_STATUS_NOT_IMPLEMENTED,
  COAP_STATUS_BAD_GATEWAY,
  COAP_STATUS_SERVICE_UNAVAILABLE,
  COAP_STATUS_GATEWAY_TIMEOUT,
  COAP_STATUS_PROXYING_NOT_SUPPORTED,
};

// content type possibili 
/* 
 * COAP CONTENT TYPE
   +------------------------------+-----+-----------------------------+
   | Media type                   | Id. | Reference                   |
   +------------------------------+-----+-----------------------------+
   | text/plain; charset=utf-8    |   0 |                             |
   | text/xml; charset=utf-8      |   1 |                             |
   | text/csv; charset=utf-8      |   2 |                             |
   | text/html; charset=utf-8     |   3 |                             |
   | application/link-format      |  40 | [I-D.ietf-core-link-format] |
   | application/xml              |  41 |                             |
   | application/octet-stream     |  42 |                             |
   | application/rdf+xml          |  43 |                             |
   | application/soap+xml         |  44 |                             |
   | application/atom+xml         |  45 |                             |
   | application/xmpp+xml         |  46 |                             |
   | application/exi              |  47 | [EXIMIME]                   |
   | application/fastinfoset      |  48 |                             |
   | application/soap+fastinfoset |  49 |                             |
   | application/json             |  50 |                             |
   | application/x-obix-binary    |  51 | [OBIX1.1]                   |
   +------------------------------+-----+-----------------------------+
*/
enum coap_contenttype_values {
  COAP_CONTENTTYPE_HTML = 3,
  COAP_CONTENTTYPE_EXI = 47,
  COAP_CONTENTTYPE_PLAIN = 0
};

