/* Copyright (c) 2009 The Regents of the Signet Group.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 *  Authors:
 *  Angelo P. Castellani <castellani@dei.unipd.it>
 *  Mattia Gheda <ghedamat@dei.unipd.it>
 */

#include "CoAP.h"

/**
 * \brief Recieves a coap_pkt_t and returns a coap_msg_t parsed
 *
 * Please note that all coap_msg_t pointer have
 * to be initialized by the caller
 */
void parse_packet( uint8_t * buf,uint32_t buflen, coap_msg_t * msg){
  coap_char_t *p, *u,*v;
  uint8_t oc;
  uint8_t lastdelta = 0;
  coap_header_t * h;

  p = (coap_char_t *)buf;
  h = (coap_header_t *)buf;

  msg->type = h->t;
  msg->code = h->code;

#ifdef i386
  msg->mid = ntohs(h->mid);
#else
  msg->mid = h->mid; 
#endif
  msg->format = COAP_CONTENTTYPE_PLAIN;

  oc = h->oc;

  p+= sizeof(coap_header_t);

  msg->token.len = 0;
  v = msg->token.val;
  //TDEBUG("oc %d",oc);
  while (oc > 0){

    coap_char_t delta;
    coap_char_t len;
    
    delta = ((coap_option_header_t* )p)->od + lastdelta;
    len = ((coap_option_header_t* )p)->len;
    p+=sizeof(coap_option_header_t); // p++;
    if (len == 0xf) // len = 15
      len += *(coap_char_t*)p++;
    
    //TDEBUG("delta %d",delta);
    //TDEBUG("len %d",len);
    switch(delta){
      case COAP_OPTION_CONTENT_TYPE:
        msg->format = *(coap_char_t*)p++;
      break;
      
      case COAP_OPTION_URI_PATH:{ // see 3.2.4 ! dubbi?
        //TDEBUG("URI OPT");
        u = msg->uri;
        // initialize uri value
        memset(msg->uri,0,COAP_MAX_URI_LEN);
        *(u++) = '/';
        while (len--) { *(u++) = *(p++); }
        *(u++) ='\0';
      }

      break;
      case COAP_OPTION_TOKEN:{
        //TDEBUG("TOKEN OPT len %d",len);
// TODO: check that len <= 8
        msg->token.len = len;
        while (len--) { *(v++) = *(p++); }
      }

      break;
      
   
      default:
        p+=len;
      break;
    }
    
    lastdelta = delta;
    oc--;
  }

  msg->payload->len = buflen - (p-buf);

  // msg->payload have to be initialized by caller
  memcpy(msg->payload->data, p, msg->payload->len);

}

/**
 * \brief Recieves a coap_msg_t and returns a coap packet ready to be sent
 */

void write_packet( coap_packet_t *pkt, coap_msg_t * msg){
  coap_char_t * p;
  uint8_t oc = 0;
  uint8_t lastdelta = 0;
  coap_header_t * h;
  coap_char_t len;
  coap_char_t* u = msg->uri;


  p = (coap_char_t*)pkt->buf;
  h = (coap_header_t*)pkt->buf;

  
  p+=sizeof(coap_header_t);
   
  h->ver = 1;
#ifdef i386
  h->mid = htons(msg->mid);
#else
  h->mid = msg->mid;
#endif

  /*
   * set CONTENT TYPE OPTION : MANDATORY
   */  

  if(msg->payload->len){
    ((coap_option_header_t* )p)->od = COAP_OPTION_CONTENT_TYPE;
    ((coap_option_header_t* )p)->len = 1;
    p+=sizeof(coap_option_header_t); // p++;
    *(coap_char_t*)p++ = msg->format;
    lastdelta=COAP_OPTION_CONTENT_TYPE;
    oc++;
  }

  /*
   * set BASE HEADER
   */

  h->t = msg->type;
  h->code = msg->code;

  /*
   * set URI OPTION
   */

  if (*u == '/')
    u++;

  len = strlen((const char *)u); // la devo conoscere prima ( il campo len putroppo sono 4 o 12 bits)
  if (len) {
    oc++;
    ((coap_option_header_t* )p)->od = COAP_OPTION_URI_PATH-lastdelta;
    if (len <0xf){
      ((coap_option_header_t* )p)->len = len;
      p+=sizeof(coap_option_header_t); // p++;
    }else{
      ((coap_option_header_t* )p)->len = 0xf;
      p+=sizeof(coap_option_header_t); // p++;
      *(coap_char_t*)p++ = len - 0xf;
    }
    lastdelta = COAP_OPTION_URI_PATH; // usless per ora, servira forse con piu' opzioni
      
    while (*u) { *(p++) = *(u++);}
  }

  /*
   * SET TOKEN OPTION
   */
  len = msg->token.len;
  u = msg->token.val;
  if (len){
    oc++;
    ((coap_option_header_t* )p)->od = COAP_OPTION_TOKEN-lastdelta;
    ((coap_option_header_t* )p)->len = len;
    p+=sizeof(coap_option_header_t); // p++;
    lastdelta = COAP_OPTION_TOKEN;
    while (len--) { *(p++) = *(u++);}
  }
  
  h->oc = oc; 
  /*
   * set the payload
   */
  if (msg->payload->len) {
    memcpy(p, msg->payload->data, msg->payload->len);
    p += msg->payload->len;
  }

  pkt->len = p - pkt->buf;
}

